/*
 * This file is part of SEMP.
 *
 * SEMP is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * SEMP is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with SEMP. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright 2009, 2010, 2011, 2012 Alpen-Adria-Universitaet Klagenfurt, Markus Waltl.
 */

#include "ColorCalculator.h"

#define JUMPPIXELY	2 // take every second row
#define JUMPPIXELX	4 // take every 5th pixel // note: because of stride the definition is: BGRBGRBGR <- 9 pixel (in video they are three)

/**
Method for calculating the color of a frame using MMX
\param pBuffer data from the current frame
\param lWidth width of the frame
\param lHeight height of the frame
\param rgb color information extracted from the frame
*/
void ColorCalculator::CalculateColorMMX(unsigned char* pBuffer, long lWidth, long lHeight, int rgb[][3])
{
	try {
		// initialize always an array of 4 ignoring if MMX is used or not
		int r_left = 0;
		int g_left = 0;
		int b_left = 0;
		int r_middle = 0;
		int g_middle = 0;
		int b_middle = 0;
		int r_right = 0;
		int g_right = 0;
		int b_right = 0;
		int cR = 0;
		int cB = 0;
		int cG = 0;
		unsigned int pixelCountLeft = 0;
		unsigned int pixelCountMiddle = 0;
		unsigned int pixelCountRight = 0;
		
		int calculatedHeight = lHeight/JUMPPIXELY;
		int screenSplitter = 0;

		// use MMX-SIMD for color calculation
		int stride = lWidth * 4; // 32bit
		
		for (long i = 0; i < lHeight; i = i + JUMPPIXELY) // i = i + 2
		{
			for (long j = 0; j < stride; j = j + (JUMPPIXELX*4) + 4) // j = j + 20
			{
				unsigned int currentX = (stride * i)+j;
				// MMX-SIMD code
				__asm
				{
					push eax
					push ebx
					
					pxor mm0,mm0
					pxor mm1,mm1
					pxor mm2,mm2
					pxor mm3,mm3
					pxor mm4,mm4

					mov eax,pBuffer
					add eax,currentX
					movd mm1, dword ptr [eax]
					add eax,4
					movd mm2, dword ptr [eax]
					add eax,4
					movd mm3, dword ptr [eax]
					add eax,4
					movd mm4, dword ptr [eax]

					punpcklbw mm1,mm0
					punpcklbw mm2,mm0
					punpcklbw mm3,mm0
					punpcklbw mm4,mm0

					paddw mm1,mm2
					paddw mm1,mm3
					paddw mm1,mm4

					movd ebx,mm1
					and ebx,0xFFFF
					mov cB,ebx

					movd ebx,mm1
					shr ebx,16
					and ebx,0xFFFF
					mov cG,ebx

					psrlq mm1,32
					movd ebx,mm1
					and ebx,0xFFFF
					mov cR,ebx
					
					pop ebx
					pop eax
					emms 
				}

				if (j <= (stride/3)){
					r_left += cR;
					g_left += cG;
					b_left += cB;
				
				}else if(j > (stride/3) && j <= ((stride/3) << 1)){
					r_middle += cR;
					g_middle += cG;
					b_middle += cB;
				
				}else if(j > ((stride/3) << 1)){
					r_right += cR;
					g_right += cG;
					b_right += cB;
					
				}
			}
		}

		// adjust screen splitter because of possible jumping
		screenSplitter = (stride/3)/((JUMPPIXELX+1));

		pixelCountLeft = calculatedHeight*screenSplitter;
		pixelCountMiddle = calculatedHeight*screenSplitter;
		pixelCountRight = calculatedHeight*screenSplitter;
		
		rgb[0][0] = r_left / pixelCountLeft;
		rgb[0][1] = g_left / pixelCountLeft;
		rgb[0][2] = b_left / pixelCountLeft;

		rgb[1][0] = r_middle / pixelCountMiddle;
		rgb[1][1] = g_middle / pixelCountMiddle;
		rgb[1][2] = b_middle / pixelCountMiddle;

		rgb[2][0] = r_right / pixelCountRight;
		rgb[2][1] = g_right / pixelCountRight;
		rgb[2][2] = b_right / pixelCountRight;
	} catch (...) {} // catch all exception because it can happen that the buffer is not filled or at an incorrect position!
}

/**
Method for calculating the color of a frame
\param pBuffer data from the current frame
\param lWidth width of the frame
\param lHeight height of the frame
\param rgb color information extracted from the frame
*/
void ColorCalculator::CalculateColor(unsigned char* pBuffer, long lWidth, long lHeight, int rgb[][3])
{
	try 
	{
		int r_left = 0;
		int g_left = 0;
		int b_left = 0;
		int r_middle = 0;
		int g_middle = 0;
		int b_middle = 0;
		int r_right = 0;
		int g_right = 0;
		int b_right = 0;
		int cR = 0;
		int cB = 0;
		int cG = 0;
		unsigned int pixelCountLeft = 0;
		unsigned int pixelCountMiddle = 0;
		unsigned int pixelCountRight = 0;
		long stride = lWidth * 3; // 24bit/8
		int screenSplitter = 0;
		int calculatedHeight = lHeight/JUMPPIXELY;

		//012 345 678 901 234 567 890 123
		//BGR BGR BGR BGR BGR BGR BGR BGR
		// 1   2   3   4   5   6   7   8
			
		for (long i = 0; i < lHeight; i = i + JUMPPIXELY) // i = i + 2
		{
			long pos = (stride * i);
			for (long j = 0; j < stride; j = j + (JUMPPIXELX*3)) // j = j + 15
			{				
				cB = pBuffer[(pos + j)];
				j++;
				cG = pBuffer[(pos + j)];
				j++;
				cR = pBuffer[(pos + j)];
				j++;

				if (j <= lWidth)
				{
					r_left += cR;
					g_left += cG;
					b_left += cB;
				}
				else if (j > lWidth && j <= (lWidth << 1))
				{
					r_middle += cR;
					g_middle += cG;
					b_middle += cB;
				}
				else if (j > (lWidth << 1))
				{
					r_right += cR;
					g_right += cG;
					b_right += cB;
				}
			}
		}

		// adjust screen splitter because of possible jumping
		screenSplitter = (unsigned int)(lWidth/(JUMPPIXELX+1))/3;				

		pixelCountLeft = calculatedHeight*screenSplitter;
		pixelCountMiddle = calculatedHeight*screenSplitter;
		pixelCountRight = calculatedHeight*screenSplitter;
				
		rgb[0][0] = r_left / pixelCountLeft;
		rgb[0][1] = g_left / pixelCountLeft;
		rgb[0][2] = b_left / pixelCountLeft;

		rgb[1][0] = r_middle / pixelCountMiddle;
		rgb[1][1] = g_middle / pixelCountMiddle;
		rgb[1][2] = b_middle / pixelCountMiddle;

		rgb[2][0] = r_right / pixelCountRight;
		rgb[2][1] = g_right / pixelCountRight;
		rgb[2][2] = b_right / pixelCountRight;

		// look:
		/*
			[0] = left
			[1] = middle
			[2] = right
			[0][0] = left_r
			[0][1] = middle_r
			[0][2] = right_r
		*/
	} catch (...) { } // catch all exception because it can happen that the buffer is not filled or an incorrect position!
}
